#pragma once

#include "mathutil/CVector.h"

#include "transfunc/TransferFunction.h"
#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "app/globaldefs.h"

class COctree;
class CBlock;

class CGPUSlicer
{
public:

	CGPUSlicer();
	~CGPUSlicer(void);
	void render(); 
	void init();
	void initOctree(int sX, int sY, int sZ);
	inline void showOctree(bool on) {m_bShowOctree = on;}
	inline void showTF(bool on) {m_bDisplayTF = on;}
	inline void setWireframe(bool on) {m_bWireframe = on;}
	inline void setMin(VTYPE nMin) {m_nMin = nMin;}
	inline void setMax(VTYPE nMax) {m_nMax = nMax;}
	inline bool handleButtonEvent(int button, int state, int x, int y);
	inline bool handleMoveEvent(int x, int y);

protected:
	GLuint   m_nTextureId;
	
	CBlock   *m_pBlock;
	COctree  *m_pOctree;	
	CTransferFunction *m_pTransferFunction;
	unsigned char *m_pVolume;

	bool m_bShowOctree;
	bool m_bDisplayTF;
	bool m_bWireframe;

	unsigned int m_nDisplayListBase;
	unsigned short *m_pListIndices;
	static bool s_bInit;

	CGprogram m_vertexProgram, m_fragmentProgram;
	CGprogram m_fragmentProgramWF;
	void loadPrograms();
	void loadTextures();
	void loadVolume();
	void initPrograms();
	void initDisplayLists();

	int m_nMin, m_nMax;
	int m_nMaxLevelX, m_nMaxLevelY, m_nMaxLevelZ;
	
	CVector m_pVertices[8];
	double m_dScaleFactorX;
	double m_dScaleFactorY;
	double m_dScaleFactorZ;
	//int m_nNumSlices;
	double m_dSliceDist;
};


bool CGPUSlicer::handleButtonEvent(int button, int state, int x, int y) {
	if (m_bDisplayTF) {
		return m_pTransferFunction->handleButtonEvent(button, state, x, y);
	} 
	return false;
}

bool CGPUSlicer::handleMoveEvent(int x, int y) {
	if (m_bDisplayTF) {
		return m_pTransferFunction->handleMoveEvent(x,y);
	} 
	return false;
}