#include "Octree.h"

#include "Block.h"
#include <math.h>
#include <assert.h>
#include <stdio.h>
//#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>
//#undef GL_GLEXT_PROTOTYPES

int COctree::s_nFrontVertex = 0;

double COctree::s_dSliceInc;

VTYPE		COctree::s_nMinValue = 0;
VTYPE		COctree::s_nMaxValue = VTYPE_MAX;
CVector		COctree::s_vecView;
CVector		COctree::s_vecCamPos;
CGprogram	COctree::s_vertexProgram = 0;
int			COctree::s_nMaxNumSlices = 0;
GLuint		COctree::s_nVertexBufferId = 0;
GLuint		COctree::s_nIndexBufferId = 0;
bool        COctree::s_bInit = false;
unsigned short *COctree::s_pIndexBuffer = NULL;

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

PFNGLBINDBUFFERPROC		glBindBufferARB		= NULL;
PFNGLDELETEBUFFERSPROC	glDeleteBuffersARB	= NULL;
PFNGLGENBUFFERSPROC		glGenBuffersARB		= NULL;
PFNGLISBUFFERPROC		glIsBufferARB		= NULL;
PFNGLBUFFERDATAPROC		glBufferDataARB		= NULL;
PFNGLACTIVETEXTUREPROC  glActiveTextureARB  = NULL;
PFNGLCLIENTACTIVETEXTUREPROC glClientActiveTextureARB = NULL;
PFNGLPRIMITIVERESTARTINDEXNVPROC glPrimitiveRestartIndexNV = NULL;

COctree::COctree(int nNumLevelsX, int nLevelX, 
				 int nNumLevelsY, int nLevelY, 
				 int nNumLevelsZ, int nLevelZ, 
				 double xMin, double xMax, 
				 double yMin, double yMax, double zMin, double zMax,
				 VTYPE *pVolume, int nSizeX, int nSizeY, int nSizeZ) 
{

	m_bIsLeaf = ((nLevelX == nNumLevelsX) && (nLevelY == nNumLevelsY) && (nLevelZ == nNumLevelsZ));

	//CVector pVertices[8];
	double  xNewMin,  xNewMax, yNewMin, yNewMax, zNewMin, zNewMax;

	m_pVertices[0] = CVector(xMin,yMin,zMin,1.0);
	m_pVertices[1] = CVector(xMax,yMin,zMin,1.0);
	m_pVertices[2] = CVector(xMin,yMax,zMin,1.0);
	m_pVertices[3] = CVector(xMax,yMax,zMin,1.0);

	m_pVertices[4] = CVector(xMin,yMin,zMax, 1.0);
	m_pVertices[5] = CVector(xMax,yMin,zMax, 1.0);
	m_pVertices[6] = CVector(xMin,yMax,zMax, 1.0);
	m_pVertices[7] = CVector(xMax,yMax,zMax, 1.0);

	m_nMinValue=255;
	m_nMaxValue=0;
	int nChild = 0;

	for(int i = 0; i < 8; ++i) {
		m_pChildren[i] = NULL;
	}

	bool bZSubdiv = (nLevelZ != nNumLevelsZ);
	bool bYSubdiv = (nLevelY != nNumLevelsY);
	bool bXSubdiv = (nLevelX != nNumLevelsX);

	if (m_bIsLeaf) {

		int nStartU = int(double(nSizeX)*((xMin+1.0)/2.0))-1;
		int nEndU   = int(double(nSizeX)*((xMax+1.0)/2.0))+1;
		int nStartV = int(double(nSizeY)*((yMin+1.0)/2.0))-1;
		int nEndV   = int(double(nSizeY)*((yMax+1.0)/2.0))+1;
		int nStartW = int(double(nSizeZ)*((zMin+1.0)/2.0))-1;
		int nEndW   = int(double(nSizeZ)*((zMax+1.0)/2.0))+1;
		if (nStartU < 0) nStartU = 0;
		if (nStartV < 0) nStartV = 0;
		if (nStartW < 0) nStartW = 0;
		if (nEndU > nSizeX-1) nEndU = nSizeX-1;
		if (nEndV > nSizeY-1) nEndV = nSizeY-1;
		if (nEndW > nSizeZ-1) nEndW = nSizeZ-1;

		VTYPE nMin = VTYPE_MAX;
		VTYPE nMax = 0;

		int u,v,w;
		for (w = nStartW; w < nEndW; w++) {
			for(v = nStartV; v < nEndV; v++) {
				for(u = nStartU; u < nEndU; u++) {
					VTYPE value = pVolume[ (w * nSizeY + v) * nSizeX + u];
					if (value < nMin) nMin = value;
					if (value > nMax) nMax = value;
				}
			}
		}

		m_pLeave = new CBlock();
		m_pLeave->setVertices(m_pVertices);
		m_pLeave->setMin(nMin);
		m_pLeave->setMax(nMax);

		m_nMaxValue = nMax;
		m_nMinValue = nMin;

	} else {

		m_pLeave = NULL;
		bool bValid;

		for(int z = 0; z < 2; ++z) {
			bValid = true;
			if (bZSubdiv) {
				zNewMin = (z==0)? zMin : (zMin + zMax)/2.0;
				zNewMax = (z==1)? zMax : (zMin + zMax)/2.0;
			} else {
				zNewMin = zMin;
				zNewMax = zMax;
			}
			if ((z > 0) && (bZSubdiv == false)) bValid = false;

			for(int y = 0; y < 2; ++y) {
				if (bYSubdiv) {
					yNewMin = (y==0)? yMin : (yMin + yMax)/2.0;
					yNewMax = (y==1)? yMax : (yMin + yMax)/2.0;
				} else {
					yNewMin = yMin;
					yNewMax = yMax;
				}
				if ((y > 0) && (bYSubdiv == false)) bValid = false;

				for(int x = 0; x < 2; ++x) {
					if (bXSubdiv) {
						xNewMin = (x==0)? xMin : (xMin + xMax)/2.0;
						xNewMax = (x==1)? xMax : (xMin + xMax)/2.0;
					} else {
						xNewMin = xMin;
						xNewMax = xMax;
					}
					if ((x > 0) && (bXSubdiv == false)) bValid = false;

					if (bValid) {

						m_pChildren[nChild] = new COctree(
							nNumLevelsX,(bXSubdiv)?nLevelX+1:nLevelX,
							nNumLevelsY,(bYSubdiv)?nLevelY+1:nLevelY,
							nNumLevelsZ,(bZSubdiv)?nLevelZ+1:nLevelZ,
							xNewMin,xNewMax,yNewMin,yNewMax,zNewMin,zNewMax,
							pVolume, nSizeX, nSizeY, nSizeZ);

					} // if (bValid)

					nChild++;

				} // for x
			} // for y
		} 

		m_nMinValue = VTYPE_MAX;
		m_nMaxValue = 0;		
		for(int i = 0; i < 8; ++i) {
			COctree *pOctree = m_pChildren[i];
			if (pOctree != NULL) {
				VTYPE nMin = m_pChildren[i]->getMin();
				VTYPE nMax = m_pChildren[i]->getMax();
				if (nMin < m_nMinValue) m_nMinValue = nMin;
				if (nMax > m_nMaxValue) m_nMaxValue = nMax;
			}
		}
	}
}

void COctree::prepareForRendering(int nFrontVertex, int nMaxNumSlices,
								  double dSliceInc, VTYPE nMinValue, VTYPE nMaxValue, const CVector &camPos, 
								  const CVector &viewVec, CGprogram vertexProgram) 
{
	if (s_bInit == false) {
		glBindBufferARB			 = (PFNGLBINDBUFFERPROC)			wglGetProcAddress("glBindBufferARB");
		glDeleteBuffersARB		 = (PFNGLDELETEBUFFERSPROC)			wglGetProcAddress("glDeleteBuffersARB");
		glGenBuffersARB			 = (PFNGLGENBUFFERSPROC)			wglGetProcAddress("glGenBuffersARB");
		glIsBufferARB			 = (PFNGLISBUFFERPROC)				wglGetProcAddress("glIsBufferARB");
		glBufferDataARB			 = (PFNGLBUFFERDATAPROC)			wglGetProcAddress("glBufferDataARB");
		glActiveTextureARB       = (PFNGLACTIVETEXTUREPROC)			wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREPROC)	wglGetProcAddress("glClientActiveTextureARB");
		glPrimitiveRestartIndexNV  = (PFNGLPRIMITIVERESTARTINDEXNVPROC) wglGetProcAddress("glPrimitiveRestartIndexNV");
	
		s_bInit = true;
	}

	s_nFrontVertex = nFrontVertex;
	s_nMaxValue = nMaxValue;
	s_nMinValue = nMinValue;
	s_vecCamPos = camPos;
	s_vecView = viewVec;
	s_vertexProgram = vertexProgram;
	s_dSliceInc = dSliceInc;

	if (s_nMaxNumSlices < nMaxNumSlices) {
		s_nMaxNumSlices = nMaxNumSlices;
		float *pVertices = new float[12*nMaxNumSlices];
		if (NULL != s_pIndexBuffer) delete [] s_pIndexBuffer;
		s_pIndexBuffer = new unsigned short[7*nMaxNumSlices];
	
		glEnableClientState(GL_PRIMITIVE_RESTART_NV);
		glPrimitiveRestartIndexNV(0xFFFF);
		int nPrimitiveRestart;
		glGetIntegerv(GL_PRIMITIVE_RESTART_INDEX_NV, &nPrimitiveRestart);

		int vidx = 0;
		int iidx = 0;
		int vcount = 0;
		for(int s = 0; s < s_nMaxNumSlices; ++s) {
			for(int i = 0; i < 6; ++i) {
				pVertices[vidx++] = float(i);
				pVertices[vidx++] = float(s_nMaxNumSlices-1-s);
				s_pIndexBuffer[iidx++] = vcount++;
				if (i == 5) {
					s_pIndexBuffer[iidx++] = nPrimitiveRestart;
				}
			}
		}

		if (s_nVertexBufferId == 0) {
			glGenBuffersARB(1, &s_nVertexBufferId);
		}
		
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, s_nVertexBufferId);
		printf("Uploading Vertex Buffer!");
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, 12*nMaxNumSlices*sizeof(float), pVertices, GL_STATIC_DRAW_ARB);

		delete [] pVertices;

		// create an index buffer		
		if (s_nIndexBufferId == 0) {
			glGenBuffersARB(1, &s_nIndexBufferId);
		}

		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,s_nIndexBufferId);
		glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,7*nMaxNumSlices*sizeof(GLushort),s_pIndexBuffer,GL_STATIC_DRAW_ARB);

	} else {
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, s_nVertexBufferId);
	}
}

void COctree::render(bool bBoundingBox) 
{

	if (s_bInit == false) {
		glBindBufferARB			 = (PFNGLBINDBUFFERPROC)			wglGetProcAddress("glBindBufferARB");
		glDeleteBuffersARB		 = (PFNGLDELETEBUFFERSPROC)			wglGetProcAddress("glDeleteBuffersARB");
		glGenBuffersARB			 = (PFNGLGENBUFFERSPROC)			wglGetProcAddress("glGenBuffersARB");
		glIsBufferARB			 = (PFNGLISBUFFERPROC)				wglGetProcAddress("glIsBufferARB");
		glBufferDataARB			 = (PFNGLBUFFERDATAPROC)			wglGetProcAddress("glBufferDataARB");
		glActiveTextureARB       = (PFNGLACTIVETEXTUREPROC)			wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREPROC)	wglGetProcAddress("glClientActiveTextureARB");
		
		s_bInit = true;
	}

	if (s_nMaxValue < m_nMinValue) {
		return;
	}
	if (s_nMinValue > m_nMaxValue) {
		return;
	}
						 
	const int nSequence[8][8] = {
		{7,3,5,6,1,2,4,0},
		{6,2,4,7,0,3,5,1},
		{5,1,4,7,0,3,6,2},
		{4,0,5,6,1,2,7,3},
		{3,1,2,7,0,5,6,4},
		{2,0,3,6,1,4,7,5},
		{1,0,3,5,2,4,7,6},
		{0,1,2,4,3,5,6,7},
	};

	int nMinIdx = 0;
	double dMinDist = (s_vecCamPos - m_pVertices[0]).getLength();

	double dDist;
	for(int v = 1; v < 8; v++) {
		dDist = (s_vecCamPos - m_pVertices[v]).getLength();
		if (dDist < dMinDist) {
			dMinDist = dDist;
			nMinIdx = v;
		}
	}
	
	if (m_bIsLeaf) {

		if (bBoundingBox){
			m_pLeave->renderBoundingBox(nMinIdx);
		} else {

			m_pLeave->activateForRendering(s_vertexProgram);

			double dStartDist   = s_vecView * m_pLeave->m_pVertices[nSequence[s_nFrontVertex][0]];
			double dEndDist     = s_vecView * m_pLeave->m_pVertices[s_nFrontVertex];
			double dS = ceil((dStartDist)/s_dSliceInc);
			dStartDist = dS * s_dSliceInc;

			int nNumSlices = int((dEndDist-dStartDist)/s_dSliceInc)+1; 
			cgGLSetParameter1d(cgGetNamedParameter(s_vertexProgram,"dPlaneStart"),dStartDist); 

			int vertexOffset = (s_nMaxNumSlices-nNumSlices)*12*sizeof(float);
			glEnableClientState(GL_VERTEX_ARRAY); 
			glVertexPointer(2, GL_FLOAT, 0, BUFFER_OFFSET(vertexOffset));
			glDrawElements(GL_TRIANGLE_FAN, (nNumSlices)*7, GL_UNSIGNED_SHORT, BUFFER_OFFSET(0));
		}


	} else {
		for(int i = 0; i < 8; ++i) {
			COctree *pOctree = m_pChildren[nSequence[nMinIdx][i]];
			if (pOctree != NULL) pOctree->render(bBoundingBox);
		}
	}
}
